"""
This module implements:

 * command class  -- base class for all native modules
 * swrapper class -- result of a module's work

"""

# 	Copyright (c) 2005-2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from cxutil import flags,pmask,cmask
import types

class command(object):
	"""
	Template command class
	"""
	comment = None
	doc = None
	mask = None
	modules = None
	children = None
	_children = None
	acquire = None
	export = None
	parameters = None
	priority = None
	subscribe = None
	catch = None
	name = None
	template = None
	fbase = None
	signal_keys = None

	def __init__(self):

		self.cfilter = None
		self.bus = None

		## the module flags
		if not self.mask:
			self.mask = 0
		self.mask |= flags.Newborn
		
		## signal mask
		if not self.signal_keys:
			self.signal_keys = {}

		## the module parameters
		if not self.parameters:
			self.parameters = {}

		## define modules...
		if not self.modules:
			self.modules = []

		## define children...
		if not self.children:
			self.children = []
		if not self._children:
			self._children = {}

		## aquire modules...
		if not self.acquire:
			self.acquire = []

		## export modules...
		if not self.export:
			self.export = []

		## subscribe list
		if not self.subscribe:
			self.subscribe = []

		## catch signals
		if not self.catch:
			self.catch = []

		## comment for the node
		if not self.comment:
			self.comment = []

		## class factory
		if not self.template:
			self.template = False

		## docstring
		if not self.doc:
			self.doc = ""

		## set custom name for the module
		if not self.name:
			self.name = ""

		## set custom child modules fbase path
		if not self.fbase:
			self.fbase = ""

		## priority
		if not self.priority:
			self.priority = 0x1000

##	TODO: create message and data bus for command class
##	TODO: think about asynchronous model (upd: threads?)
	def __call__(self,opts):
		return self.pre(opts)

	def condition(self,condition):
		return True

	def fetch(self,opts,key,default):
		'''
		'''
		if opts.has_key(key):
			return opts[key]
		else:
			return default

	def create_condition(self,opts):
		'''
		Get condition variables from command line and store it
		'''

		###
		# FIXME and remove condition at all
		###
		return None

	def set(self,opts,parameter,name,value,key):
		if parameter.has_key("format"):
			try:
				value = parameter["format"].substitute(value=value)
			except:
				pass

		if not value:
			if parameter.has_key("default"):
				value = parameter["default"]
			else:
				return None

		if parameter.has_key("mask"):
			if parameter["mask"] & pmask.List:
				if not opts.has_key(name):
					opts[name] = []
				if \
					(type(value) == types.ListType) or \
					(type(value) == types.TupleType):
					opts[name].extend(value)
				else:
					opts[name].append(value)
				return None

		if parameter.has_key("dict"):
			if not opts.has_key(name):
				opts[name] = {}
			if parameter.has_key("dict"):
				if parameter["dict"] == "key":
					opts[name][value] = True
					return value
				elif (parameter["dict"] == "value"):
					if key:
						opts[name][key] = value
					else:
						print "DEBUG: no key for value %s" % (value)
			return None
		
		opts[name] = value
		return None

	def list(self):
		return self.parameters

	def signal(self,sig,data):
		self.bus.put(
			"state",
			{
				"map": "signal",
				"signal": sig,
				"data": data,
			}
		)

	def pass_filter(self,data):
		if self.cfilter is None:
			return True

		if (type(data) == type(self.cfilter) == types.StringType) and (data == self.cfilter):
			return True

		try:
			for i in self.cfilter.keys():
				if data[i] in self.cfilter[i]:
					return True
		except:
			pass

		return False

	def add_signal_filter(self,key,value=None):
		if isinstance(value,type(None)):
			self.cfilter = key
		else:
			if type(self.cfilter) != types.DictType:
				self.cfilter = {}
			if self.cfilter.has_key(key):
				if not value in self.cfilter[key]:
					self.cfilter[key].append(value)
			else:
				self.cfilter[key] = [value,]

	def del_signal_filter(self,key=None,value=None):
		if (type(self.cfilter) == types.DictType) and (key):
			if value:
				try:
					self.cfilter[key].remove(value)
				except:
					pass
			else:
				try:
					del self.cfilter[key]
				except:
					pass
		else:
			self.cfilter = None


	###
	# In the order of calls:
	###

	def pre(self,opts):
		'''
		Called from state/core.py:state2.parse() if flags.Immediate or bypass()
		Called from state/core.py:state2.commit()

		Get options from command line and store it in the locals
		'''
		result = swrapper()
		if self.mask & flags.Once:
			if opts.already:
				return result

		for i in opts.opts.deprecated:
			result.add_command("message","deprecated parameter definition: %s, %s: %s\n" % (opts.fn,i[0],i[1]))

		key = None
		keys = opts.opts.parameters.keys()
		keys.sort()
		for i in keys:
			if self.parameters[i].has_key("name"):
				key = self.set(opts.envl,self.parameters[i],self.parameters[i]["name"],opts.opts.parameters[i],key)
				if self.mask & flags.Upload:
					self.set(opts.rest_tp,self.parameters[i],self.parameters[i]["name"],opts.opts.parameters[i],key)
			else:
				key = self.set(opts.envl,self.parameters[i],i,opts.opts.parameters[i],key)
				if self.mask & flags.Upload:
					self.set(opts.rest_tp,self.parameters[i],i,opts.opts.parameters[i],key)

		return result

	def immediate(self,opts):
		'''
		Called from state/core.py:state2.parse() right after command input
		'''
		return swrapper()

	def init(self,opts):
		'''
		Called from state/core.py:state2.commit() at the first node commit
		'''
		return swrapper()

	def postUp(self,opts):
		'''
		Called from state/core.py:state2.commit() at every commit
		'''
		return swrapper()

	def postDown(self,opts):
		'''
		Called from state/core.py:state2.commit() at every commit
		'''
		return swrapper()
	
	def terminate(self,opts):
		'''
		Called from state/core.py:state2.commit() at the last node commit
		'''
		return swrapper()

class not_ready(command):
	def immediate(self,opts):
		'''
		Return "not implemented"
		'''
		return swrapper({"message": "Not implemented, sorry\n"})
		

class swrapper(object):
	
	__internal_ = None
	cookie = "none"
	
	def __init__(self,dct = {}):
		self.__internal_ = []
		if dct.keys():
			for (i,k) in dct.items():
				if \
					type(k) == types.TupleType or \
					type(k) == types.ListType:
					self.add_commands(i,k)
				elif \
					type(k) == types.StringType:
					self.add_command(i,k)
				else:
					pass

	def add_command(self, oid, cmd):
		self.__internal_.append({
			"service": oid,
			"command": cmd,
		})
	
	def add_commands(self, oid, list):
		for i in list:
			self.add_command(oid,i)

	def insert_commands(self, oid, list):
		old = self.__internal_
		self.__internal_ = []
		self.add_commands(oid, list)
		self.__internal_.extend(old)

	def get_commands(self):
		return self.__internal_

	def merge(self,wrappers):
		for i in wrappers:
			self.__internal_.extend(i.get_commands())

	def empty(self):
		return not len(self.__internal_)

	def __repr__(self):
		ret = ""
		for i in self.__internal_:
			ret += "%s: %s\n" % (i["service"], i["command"])
		if not ret:
			ret = "<empty>\n"
		return ret
