#!/usr/bin/env python
"""
Head module for Connexion project core
"""

# 	Copyright (c) 2005-2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import os
import sys
import traceback
## from shell import shell

from bus.dispatch import dispatcher2 as _dispatcher
from bus.monitor import monitor as _monitor
from cxcore.logger import logger as _logger
from gateway.core import server as _gateway
from gateway.core import sessionManager, proxy

from orb.core import orb as _orb
from state.core import manager
from executive.runner import runner as _runner

# commented out due to ALT specific, till it'll be solved
#
# from setproctitle import setproctitle

from getopt import getopt

import time
import threading
import tempfile

# commented out due to ALT specific, till it'll be solved
#
# setproctitle("shell server")

local_shell = False
socketG = "/".join((tempfile.gettempdir(),"socket"))
socketT = ""
socketN_a = ""
socketN_p = ""
cluster_a = "0.0.0.0"
cluster_m = []
cluster_p = 40323
cluster_name = ""
cluster_psk = None
wdir = []
prefixes = []
flag = None
logfile = None
loglevel = "error"
protodebug = False


def clean():
	try:
		os.unlink(socketG)
		if socketT:
			os.unlink(socketT)
	except:
		pass


def usage():
	print '''
Shell server: the core.

Options:
	-h		this help

	# 8<------------------------------------------------------------
	# clusterization (disabled by default)
	#
	-b <addr>	cluster manager address (default: 0.0.0.0)
	-c <port>	cluster manager port (default: 40323)
	-m <address>	join multicast group on this address
	-N <name>	use name for mDNS service instance (if set up,
			enables clusterization capabilities)
	-k <path>	path to the PSK file
	

	# 8<------------------------------------------------------------
	# debugging and logging
	#
	-d		turn on server <-> client protocol debug printout
	-D <level>	loglevel: critical, error, warning, info, debug
	-l <path>	log file to use (logs to stdout, if not specified)
	-x		run local debug shell

	# 8<------------------------------------------------------------
	# RPC and startup flags
	-f <path>	flag pipe (see docs before using!)
	-n <addr>	client interface socket addr (tcp/ip, trusted) (see -p)
	-p <port>	client interface port (tcp/ip) (see -n)
	-s <path>	client interface socket path (untrusted)
	-t <path>	client interface socket path (trusted)

	# 8<------------------------------------------------------------
	# modules
	#
	-w <prefix>:<path>	add modules directory to the tree stack
	-W <prefix>:<path>	add privileged modules directory

For details, see the documentation.
'''

try:
	(opts,left) = getopt(sys.argv[1:],"b:c:dD:hf:k:l:m:N:n:p:r:s:t:w:W:x")
except Exception,e:
	print e
	usage()
	sys.exit(255)

for (i,k) in opts:
	if i == "-b":
		cluster_a = k
	elif i == "-c":
		cluster_p = int(k)
	elif i == "-m":
		cluster_m.append(k)
	elif i == "-N":
		cluster_name = k
	elif i == "-k":
		cluster_psk = k
	elif i == "-d":
		protodebug = True
	elif i == "-D":
		loglevel = k
	elif i == "-h":
		usage()
		sys.exit(0)
	elif i == "-f":
		flag = k
	elif i == "-l":
		logfile = k
	elif i == "-n":
		socketN_a = k
	elif i == "-p":
		socketN_p = int(k)
	elif i == "-s":
		socketG = k
	elif i == "-t":
		socketT = k
	elif i == "-w":
		k = k.split(":")
		k.append(0)
		wdir.append(k)
		prefixes.append(k[0])
	elif i == "-W":
		k = k.split(":")
		k.append(1)
		wdir.append(k)
		prefixes.append(k[0])
	elif i == "-x":
		local_shell = True

###
#
# python 2.4 logger module requires stderr to be opened, if no logfile specified
#
###
if (not logfile) and (not local_shell):
	print """
Please specify logfile before run connexion in the background.
Use -l to point a logfile or -x to run in the foreground.
"""
	sys.exit(255)

class CX_LoggerStartException(Exception):
	pass

class CX_LoggedStartException(Exception):
	pass

def init_threads():

	try:
		## dispatcher thread
		dispatcher = _dispatcher(
			name = "dispatcher2",
			sname = cluster_name,
			address = cluster_a,
			port = cluster_p,
			join = cluster_m,
			psk = cluster_psk
		)

		dispatcher.start()

		## only connect: we do not need to receive any messages
		local = dispatcher.connect()

	except:
		raise

	try:
		## logger thread
		logger = _logger("logger",logfile,loglevel)
		logger.bus = dispatcher.connect()
		dispatcher.subscribe(logger.bus.address,"logger")
		logger.start()
	except:
		raise CX_LoggerStartException()

	try:
		## socket monitor
		monitor = _monitor(dispatcher.connect(),"socket monitor")
		dispatcher.subscribe(monitor.bus.address,"monitor")
		monitor.setDaemon(True)
		monitor.start()

		## background runner
		runner = _runner(dispatcher.connect(),"runner for background jobs")
		dispatcher.subscribe(runner.bus.address,"runner")
		runner.setDaemon(True)
		runner.start()

		## orb thread
		orb = _orb(dispatcher.connect(),"orb",wdir)
		dispatcher.subscribe(orb.bus.address,"orb")
		orb.start()

		## state manager
		state = manager(dispatcher.connect(),"state manager",{"prefixes": prefixes})
		dispatcher.subscribe(state.bus.address,"state")
		state.start()

		## session manager
		smanager = sessionManager(dispatcher.connect(),prefixes)
		dispatcher.subscribe(smanager.bus.address,"session manager")
		smanager.start()

		## gateway thread
		gateway = _gateway(dispatcher.connect(),proxy,"XML-RPC gateway",socketG,trusted=False,debug=protodebug)
		gateway.setDaemon(True)
		gateway.start()

		## trusted gateway
		if socketT:
			trusted = _gateway(dispatcher.connect(),proxy,"XML-RPC gateway (trusted)",socketT,trusted=True,debug=protodebug)
			trusted.setDaemon(True)
			trusted.start()
	except:
		local.put("logger",("critical",traceback.format_exc()))
		local.put("dispatcher2","shutdown")
		time.sleep(1)
		raise CX_LoggedStartException()
	
	#if socketN_p:
	#	network = _gateway(dispatcher.connect(),proxy,"XML-RPC gateway (network)",(socketN_a,socketN_p),trusted=True,debug=protodebug,prefixes=prefixes)
	#	network.setDaemon(True)
	#	network.start()

	return (dispatcher,local)


def shell(bus):
	print """
Debug shell started. Autocomplete does not work. See 'help' for details
Executable file was started from %s
""" % (sys.argv[0])
	while True:

		try:
			text = raw_input("%s # " % (cluster_name))
		except KeyboardInterrupt:
			print
			text = ""
		except:
			print "quit"
			text = "quit"

		if not text:
			continue

		if text == "threads":
			threads = threading.enumerate()
			threads.sort()
			for i in threads:
				print " * %s" % (i.getName())
		elif text == "arp":
			for i in dispatcher.arp():
				print i
		elif text == "groups":
			for i in dispatcher.groups():
				print str(i) + ":"
				for k in dispatcher.subscribers(i):
					print "\t" + str(k)
		elif text == "cache":
			print bus.put2("orb",
				{
					"map": "cache"
				}
			)

		elif text == "tree":
			print bus.put2("state",
				{
					"map": "trace"
				}
			)

		elif text == "clean":
			print bus.put2("orb",
				{
					"map": "clean"
				}
			)

		elif text[:3] == "ptr":
			print bus.put2("dispatcher2",text)

		elif text[:6] == "lookup":
			print bus.put2("dispatcher2",text)

		elif text[:5] == "state":
			bus.put("dispatcher2",text)

		elif text[:5] == "level":
			bus.put("logger",text.split()[1])

		elif text == "dump":
			print bus.put2("dispatcher2","debug")

		elif text == "no dump":
			print bus.put2("dispatcher2","no debug")

		elif text == "dns":
			print bus.put2("dispatcher2","dns")

		elif text == "getacl":
			print bus.put2("orb",
				{
					"map": "getacl"
				}
			)

		elif text == "trace":
			print bus.put2("runner",
				{
					"map": "debug",
				}
			)

		elif text == "fd":
			os.system("ls -l /proc/%s/fd/" % (os.getpid()))

		elif text == "debug":
			password = raw_input("Enter password for debug connection: ")
			try:
				import rpdb2
				print "starting rpdb2"
				rpdb2.start_embedded_debugger(password)
			except:
				print "no rpdb2 found, continuing w/o debugging"

		elif text == "help":
			print '''
Available commands:

	command		(subsystem)	description
	-------		----------	-----------

	arp		(dispatcher)	inspect arp array
	cache		(orb)		inspect object cache
	clean		(orb)		clean object cache
	debug		(main)		run embedded debugger (rpdb2)
	dns		(dispatcher)	inspect mDNS database
	dump		(dispatcher)	turn dispatcher2 packet dumping on
	fd		(main)		inspect open file descriptors
	getacl		(orb)		inspect installed ACLs
	groups		(dispatcher)	inspect multicast groups
	level <level>	(logger)	set up logging level
	no dump		(dispatcher)	turn dispatcher2 packet dumping off
	threads		(main)		inspect running threads
	trace		(runner)	switch background jobs runner debug
	quit		(main)		stop the server and leave the shell (Ctrl-D)

Embedded debug shell is just very, very simple input interpreter with
no readline/autocomletion support etc.

'''

		elif text == "quit":
			bus.put("dispatcher2","shutdown")
			time.sleep(0.5)
			clean()
			sys.exit(0)

		else:
			print "syntax error, see 'help' for details"


clean()

if local_shell:
	(d,b) = init_threads()
	shell(b)
	d.join()

else:
	(r,w) = os.pipe()
	p = os.fork()
	# child
	if p == 0:
		# close file descriptors
		for i in xrange(1024):
			try:
				if i != w:
					os.close(i)
			except:
				pass
		try:
			(d,b) = init_threads()
			os.write(w,"0")
			d.join()
		except CX_LoggerStartException:
			os.write(w,"l")
		except CX_LoggedStartException:
			os.write(w,"e")
		except:
			os.write(w,"x")
			os.write(w,"%s" % (traceback.format_exc()))

		clean()
		sys.exit(0)
	# parent
	else:
		x = os.read(r,1)
		if x == "0":
			sys.exit(0)
		else:
			
			if x == "x":
				x = os.read(r,65535)
				print "Dispatcher startup error:\n8< ======================"
				print x

			elif x == "l":
				print "Logger startup error. Please check logfile permissions."

			elif x == "e":
				print "Startup error, see logfile."

			os.kill(p,15)
			time.sleep(2)
			os.kill(p,9)

			sys.exit(255)

sys.exit(1)
