"""
Shell-code execution dispatcher
"""

# 	Copyright (c) 2005-2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import types
import os
import imp
import service
import re
import mimetypes
from popen2 import Popen3
from select import poll,POLLIN,POLLERR,POLLHUP,POLLPRI

from bus import transport

from cxcore.manager import ManagerThread
from cxutil import emask
from cxutil.UID import UID

class runner(ManagerThread):
	
	bus = None
	orb_bus = None
	services = None
	map = None
	dump = None
	
	def post(self):
		self.orb_bus = self.bus.put2("dispatcher2","connect")
		self.services = {
			"exec": self.service_exec,
			"retval": self.service_retval,
			"system": self.service_system,
			"daemon": self.service_daemon,
			"message": self.service_message,
			"data": self.service_data,
			"pedit": self.service_pedit,
			"logger": self.service_logger,
		}
		self.map = {
			"perform": self.perform,
		}
		self.daemons = {}
		self.dump = False


	def service_exec(self,bus,addr,pipe,ctl,flags,commands,cookie):
		'''
		Run subprocess via fork'n'exec
		'''
		return self.service_shell_common(bus,addr,pipe,ctl,flags,commands,cookie,fc=True)

	def service_system(self,bus,addr,pipe,ctl,flags,commands,cookie):
		'''
		Run subprocess in the shell
		'''
		return self.service_shell_common(bus,addr,pipe,ctl,flags,commands,cookie,fc=False)

	def service_shell_common(self,bus,addr,pipe,ctl,flags,commands,cookie,fc):
		result = 0
		for command in commands:
			uid = UID()

			###
			# To run subprocess in the shell, one should use StringType for popen
			# object. If on provides TupleType or ListType, a subprocess will be
			# executed via fork/exec
			#
			# fc stands for `fast call` and results in the ListType command
			#
			if fc:
				command = command.split()

			inst = Popen3(command,True,-1)
			#             ^^^^^^^
			(i,k,l) = (inst.fromchild, inst.tochild, inst.childerr)
			pid = inst.pid
			#
			###
			
			## 8<----------------------------------------------------------------------------------------------
			## Start a two-way read/write stream
			x = poll()
			data = ""
			t = 100
			x.register(i)
			x.register(l)
			if pipe:
				x.register(pipe)
			if ctl:
				x.register(ctl)
			try:
				while True:
					m = x.poll(t)
					if not m:
						if (data) and (not flags & emask.Background):
							bus.put(addr,
								{
									"uid": uid,
									"cookie": cookie,
									"data": transport.encode(data).as_string(),
								}
							)
							data = ""
					for (fd,event) in m:
						if \
							(event & POLLIN) or\
							(event & POLLPRI):
							a = os.read(fd,1)
							if fd == pipe:
								k.write(a)
								k.flush()
							if fd == ctl:
								if a == 'c':
									os.kill(pid,2)
							if \
								(fd == i.fileno()) or \
								(fd == l.fileno()):
								data += a
						else:
							if fd == l.fileno():
								continue

							if (data) and (not flags & emask.Background):
								bus.put(addr,
									{
										"uid": uid,
										"cookie": cookie,
										"data": transport.encode(data).as_string(),
									}
								)
								data = ""
							raise
			except:
				x.unregister(i)
				x.unregister(l)
				if pipe:
					x.unregister(pipe)
				if ctl:
					x.unregister(ctl)
			retval = 255
			try:
				status = os.waitpid(pid,0)
				if os.WIFEXITED(status[1]):
					retval = os.WEXITSTATUS(status[1])
			except:
				pass
			
			i.close()
			k.close()
			l.close()

			if retval != 0:
				result = 1
		
			##
			## 8<----------------------------------------------------------------------------------------------
		return result

	def service_retval(self,bus,addr,pipe,ctl,flags,commands,cookie):
		return commands[0]

	def service_logger(self,bus,addr,pipe,ctl,flags,commands,cookie):
		"""
		Send strings to the logging facility
		"""

		for i in commands:
			bus.put("logger",("info",str(i)))

		return 0

	def service_message(self,bus,addr,pipe,ctl,flags,commands,cookie):
		major = "text"
		minor = "plain"
		uid = UID()
		for command in commands:
			if (command) and (not flags & emask.Background):
				bus.put(addr,{
					"uid": uid,
					"cookie": cookie,
					"data": transport.encode(command,major_type=major,minor_type=minor).as_string()
				})
		return 0

	def service_data(self,bus,addr,pipe,ctl,flags,commands,cookie):
		uid = UID()
		for command in commands:
			if (command) and (not flags & emask.Background):
				bus.put(addr,{
					"uid": uid,
					"cookie": cookie,
					"data": transport.encode(command[2],major_type=command[0],minor_type=command[1]).as_string()
				})
		return 0

	def daemon_start(self,c):
		i = c["id"]
		self.daemons[i] = {
			"condition": c["condition"],
		}
		p = os.fork()
		# child:
		if p == 0:
			# close file descriptors
			for i in xrange(1024):
				try:
					os.close(i)
				except:
					pass
			# change wd to /
			os.chdir("/")
			# exec a command
			x = c["command"].split()
			os.execvp(x[0],x)

		# parent:
		else:
			self.daemons[i]["pid"] = p

	def daemon_stop(self,i):
		pid = self.daemons[i]["pid"]
		os.kill(pid,15)
		(p,s) = os.waitpid(pid,os.WNOHANG)
		if p == 0:
			os.kill(pid,9)
			(p,s) = os.waitpid(pid,0)

	def service_daemon(self,bus,addr,pipe,ctl,flags,commands,cookie):
		retval = 0
		for c in commands:
			if c["action"] == "check":
				try:
					if (self.daemons.has_key(c["id"])):
						if c["condition"] != self.daemons[c["id"]]["condition"]:
							self.daemon_stop(c["id"])
					self.daemon_start(c)
				except:
					retval = 255
		return retval

	def service_pedit(self,bus,addr,pipe,ctl,flags,commands,cookie):
		# FIXME
		# used in routing table
		target = ""
		mode = ""
		actions = {}
		out = []
		file = ""

		for command in commands:
			if mode and target:
				actions[target] = (mode,command)
				mode = ""
				target = ""

			if command[:7] == "replace":
				target = command[7:].strip()
				mode = "replace"

			if command[:6] == "remove":
				target = command[6:].strip()
				actions[target] = ("remove",)
				target = ""

			if command[:4] == "file":
				file = command[4:].strip()

		try:
			f = open(file,"r")
			for i in f.readlines():
				s = i.strip()
				key = re.sub('[:= \t].*','',s)
				if key in actions.keys():
					'''
					FIXME: this algorithm of token removal does NOT work
						with lists like that:

						[ section ]
						key = value1
						key = value2
						...

						For such lists the algo must remove the whole
						line match, not the key match only!
					'''
					if actions[key][0] == "remove":
						pass


					if actions[key][0] == "replace":
						pref = re.sub('[^ \t].*','',i)
						sep = re.sub('%s[ \t]*([^ \t]).*','\\1' % (key),i)
						out.append("%s%s%s%s" % (pref,key,sep,actions[key][1]))

					del actions[key]
				else:
					out.append(i)
			f.close()
		except:
			pass

		for i in actions.keys():
			if actions[i][0] == "replace":
				out.append("%s\t%s" % (i,actions[i][1]))

		try:
			f = open(file,"w")
			for i in out:
				if i:
					if i[-1] != '\n':
						i += '\n'
				f.write(i)
			f.close()
		except:
			pass

		return 0

	def debug(self,opts):
		"""
		"""
		if self.dump:
			self.dump = False
			return "service debug switched OFF\n"
		else:
			self.dump = True
			return "service debug switched ON\n"

	def perform(self,opts):
		"""
		opts['addr'] - mangler's address
		opts['pipe'] - input pipe (from an interface to a service)
		opts['ctl'] - control pipe (from an interface to a service)
		opts['intermediate'] - should the call return 'retval' or not
		opts['services'] - services list
		"""
		addr = opts['addr']
		pipe = opts['pipe']
		ctl = opts['ctl']
		services = opts['services'] 
		if opts.has_key("flags"):
			flags = opts["flags"]
		else:
			flags = emask.Empty
		
		retval = 0
		m = 1

		if services:
			log = "service commands:\n8<----------------\n"
			for i in services:
				log += "%s" % (i)
			log += "8<----------------"
			self.bus.put("logger",("debug",log))


		for i in services:
			if type(i) == types.StringType:
				# FIXME: deprecate
				self.bus.put("logger",("warning","deprecated service usage: `%s'" % (i)))
				self.service_system(self.bus,addr,pipe,ctl,flags,(i,),{})
			else:
				if i:
					service = ""
					target = None
					commands = []
					for k in i.get_commands():

						if self.dump:
							k["command"] = str(k["service"]) + ": " + str(k["command"]) + "\n"
							k["service"] = "message"
							print k["command"],

						if (target) and (service != k["service"]):
							m = target(self.bus,addr,pipe,ctl,flags,commands,i.cookie)
							if m != 0:
								retval = m
							service = ""
							target = None
							commands = []

						commands.append(k["command"])

						if not service:
							service = k["service"]
							if not self.services.has_key(service):
								target = self.orb_bus.put2("orb",
									{
										"map":  "service",
										"id": service
									}
								)
							else:
								target = self.services[service]


					if target:
						m = target(self.bus,addr,pipe,ctl,flags,commands,i.cookie)
						if m != 0:
							retval = m

		if not flags & emask.Background:
			return {"retval": retval}


