# -*- coding: utf-8 -*-
"""
Command line parser; internal 'magic' commands
"""

# 	Copyright (c) 2005-2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from re import sub
from copy import copy
from command import command, swrapper
from cxutil import utils,flags
from time import time

import os
import sys
import inspect
import types
import re

from state.syntax import oparser,xSyntaxError,xSyntaxErrorNoMandatory,xSyntaxErrorNotCompleted,xParamError



class mangler(object):
	'''
	Name mangling.
	'''
	
	oid = None
	bus = None
	runner = None
	comment = None

	queue = None
	prefixes = None

	pipe_read = None
	pipe_write = None
	ctl_read = None
	ctl_write = None

	debug = None


	def __init__(self,bus,runner,oid,prefixes):
		'''
		Create an interpreter instance.
		'''

###	Create transaction tree and stack for tracing current position
		# master connection
		self.bus = bus
		# xInstance retrieval
		self.busInstance = self.bus.put2("dispatcher2","connect")
		# execution data channel
		self.busRunner = self.bus.put2("dispatcher2","connect")

		# state manager, version 2
		self.busState = self.bus.put2("dispatcher2","connect")
		self.s2node = self.busState.put2(
			"state", { "map": "getRoot" }
		)
		self.s2points = []

		self.debug = {}
		self.runner = runner
		self.oid = oid
		self.comment = []
		(self.pipe_read, self.pipe_write) = os.pipe()
		(self.ctl_read, self.ctl_write) = os.pipe()

##		check, if the bus is empty
		self.bus.drop()
		self.queue = 0

		self.prefixes = prefixes

	def input(self,creds,text):
		'''
		'''
		if text:
			os.write(self.pipe_write,text)

		result = {"data": []}


		while (self.queue > 0):
			addr = {}

			t = time()
			r = self.busRunner.get(addr,0.5)
			if r:
				result["data"].append(r)
				if r.has_key("retval"):
					self.queue -= 1

			if (time() - t) > 0.1:
				break

		result["queue"] = self.queue
		return result
		

	def interrupt(self,creds,code):
		'''
		'''
		if code:
			os.write(self.ctl_write,code)
		return True

	def parse(self,creds,cookie,text,lines = [],noout = False):
		'''
		'''

		result = {}
		error = self.s2mangle(creds,cookie,text,lines)

		if error:
			result["error"] = error

		result["prompt"] = self.s2node.fn

		return result

	def s2mangle(self, creds, cookie, text, lines = [], queue = "up", intermediate = False, callback = False):

		# remove leading and trailing spaces
		text = text.strip()
		# remove double spaces
		text = " ".join(text.split())

		#self.bus.put("logger@connexion",
		#	(
		#		"info",
		#		"text input: %s" % (text)
		#	)
		#)

		chain = self.bus.put2("orb",
			{
				"map": "get_chain",
				"creds": creds,
				"text": text,
				"prefix": self.s2node.module.prefix,
				"namespace": self.s2node.module.namespace
			}
		)

		
		if chain:
			# create a dedicated channel for modules
			self.busMods = self.bus.put2("dispatcher2","connect")
			#
			message = self.busState.put2(
				"state",
				{
					"esoteric": {
						"runner": self.runner,
						"bus": self.busMods,
						"pipe": self.pipe_read,
						"ctl": self.ctl_read,
						"mangler": self,
					},
					"map": "chain",
					"creds": creds,
					"cookie": cookie,
					"node": self.s2node,
					"chain": chain,
					"oid": self.oid,
					"comment": self.comment,
				}
			)
			# disconnect the channel
			self.bus.put("dispatcher2",{"disconnect": self.busMods.address})
			# catch shutdown
			if message == "shutdown":
				return
			#
			(s2n,message,self.comment) = message

			if s2n:
				if s2n != self.s2node:
					self.s2points.append(self.s2node)
					self.s2node = s2n
		else:
			s = swrapper()
			s.add_commands("message",["syntax error: `%s'\n" % (text),])
			message = [s,]

		if not message:
			message = [swrapper(),]


		self.busRunner.put(
			self.runner,
			{
				"map": "perform",
				"pipe": self.pipe_read,
				"ctl": self.ctl_read,
				"services": message,
			}
		)
