"""
Module cache manager
"""

# 	Copyright (c) 2005-2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import inspect
import imp
import types
import sys
import os

from cxutil.UID import UID
from cxutil import flags,utils
from copy import copy,deepcopy
from command import command

class ExceptionNoImport(Exception):
	pass

class cache(object):
	'''
	Implements class cache
	'''
	modules = None
	pathkey = None
	bus = None
	wdir = None

	def __init__(self,bus,wdir=[]):
		self.bus = bus
		self.wdir = wdir
		self.pathkey = {}
		self.modules = {}

		p = "/".join(sys.argv[0].split("/")[:-1])

		for i in wdir:
			imp.load_source(i[0],"%s/command.py" % (p))
			self.modules[i[0]] = {
				"class": cfactory(i[0]),
				"path": i[1],
				"namespace": i[0],
				"pyfqdn": i[0],
				"privileged": i[2],
			}


	def bala(self):
		result = ""
		for i in self.modules.keys():
			result += "%s\n" % (i)
		result += "\ntotal: %i entries" % (len(self.modules))
		return result

	def get(self, oid):
		cmd = {
			"class": None,
			"instance": None,
			"children": [],
			"path": "",
			"name": "",
			"pyfqdn": "",
		}
		xClass = None


		if self.modules.has_key(oid):
			# self.bus.put("logger",("debug","hit from cache: " + oid))
			cmd = self.modules[oid]
		else:

			# self.bus.put("logger",("debug","miss from cache: " + oid))
			
			base = " ".join(oid.split()[:-1])
			name = oid.split()[-1]

			###
			# Load all parent modules
			###
			if oid.find(" ") > -1:
				parent = self.get(base)
			else:
				return cmd
			
			## FIXME: replace / with os.pathsep AND test it
			## at first, let's try to get a file with the same name:
			path = parent["path"]

			listing = self.get_unknown(parent)
			if listing.has_key(name):
				xClass = listing[name]["class"]

			### 8<---------------------------------------------------------------------------------
			# Here, we have tried all ways to find a module. If we still haven't
			# got any result, there are no such module to load. "Sorry guys" and so on...
			###
			if xClass:
				## we have xClass, but is it a class? :) paranoia...
				if inspect.isclass(xClass):
					## store it in the cache
					self.load(oid,xClass,path + "/" + name)
					cmd = self.modules[oid]

		return cmd

	def get_file(self,path,pyfqdn,prefix,i):
		listing = {}
		fbase = None
		zModule = None
		zClass = None
		try:
			if i[-4:] == ".pyc":
				fbase = i[:-4]
				fqdn = pyfqdn + "." + fbase
				key = path + "/" + fbase
				if self.pathkey.has_key(key):
					zModule = self.pathkey[key]
				else:
					try:
						zModule = imp.load_compiled(fqdn, path + "/" + i)
						self.pathkey[key] = zModule
					except Exception,e:
						print "failed to load %s: %s" % (path + "/" + i,e)
						raise e
						
			elif i[-3:] == ".py":
				fbase = i[:-3]
				fqdn = pyfqdn + "." + fbase
				key = path + "/" + fbase
				if self.pathkey.has_key(key):
					zModule = self.pathkey[key]
				else:
					try:
						zModule = imp.load_source(fqdn, path + "/" + i)
						self.pathkey[key] = zModule
					except Exception,e:
						print "failed to load %s: %s" % (path + "/" + i,e)
						raise e
		except:
			pass

		try:
			if zModule:
				if hasattr(zModule,"no_import"):
					raise Exception("import denied")

				try:
					zClass = getattr(zModule,fbase)
				except:
					zClass = getattr(zModule,"Default")
			if zClass:
				if zClass.name:
					oid = prefix + " " + zClass.name
					self.load(oid,zClass,fqdn,path + "/" + fbase)
					listing[zClass.name] = self.modules[oid]
				else:
					oid = prefix + " " + fbase
					self.load(oid,zClass,fqdn,path + "/" + fbase)
					listing[fbase] = self.modules[oid]

				zClass._children = {}
				if zClass.children:
					self.get_unknown(self.modules[oid])

				for k in zClass.children:
					_oid = "%s %s" % (oid,k)
					zClass._children[_oid] = self.modules[_oid]["class"]

		except:
			pass

		return listing

	def get_unknown(self,entry):
		listing = {}
		xClass = entry["class"]
		if not xClass:
			return listing

		path = entry["path"]
		pyfqdn = entry["pyfqdn"]
		prefix = xClass.prefix

		###
		# List file modules in the dir
		###
		sys.path.append(path)
		try:
			flist = os.listdir(path)
			flist.sort()
			for i in flist:
				utils.merge(self.get_file(path,pyfqdn,prefix,i),listing)

		except:
			pass
		sys.path.pop()


		###
		# Populate .modules by template_module
		###
		if \
			(type(xClass.template) == types.ListType) or\
			(type(xClass.template) == types.TupleType):

			series = xClass.template[0]
			base = xClass.template[1]

			if not (type(xClass.modules) == types.ListType):
				xClass.modules = []

			for i in series():
				cid = str(UID())
				m = type(cid, base, i)
				setattr(xClass,cid,m)
				xClass.modules.append([m.name,cid])

		###
		# List .modules
		###
		if \
			(type(xClass.modules) == types.ListType) or\
			(type(xClass.modules) == types.TupleType):
			for i in xClass.modules:
				if type(i) == types.StringType:
					try:
						x = getattr(xClass,i)
						if x.name:
							oid = prefix + " " + i
							self.load(oid,x)
							listing[x.name] = self.modules[oid]
						else:
							oid = prefix + " " + i
							self.load(oid,x)
							listing[i] = self.modules[oid]
					except:
						pass
				elif \
					(type(i) == types.TupleType) or\
					(type(i) == types.ListType):
					try:
						x = getattr(xClass,i[1])
						oid = prefix + " " + i[0]
						self.load(oid,x)
						listing[i[0]] = self.modules[oid]
					except:
						pass


		###
		# List acquired modules
		###
		if \
			(type(xClass.acquire) == types.ListType) or\
			(type(xClass.acquire) == types.TupleType):
			for i in xClass.acquire:
				if type(i) == types.StringType:
					x = self.get_acquired(entry,i)
					if x:
						oid = prefix + " " + i
						self.load(oid,x)
						listing[i] = self.modules[oid]
				elif \
					(type(i) == types.TupleType) or\
					(type(i) == types.ListType):
					x = self.get_acquired(entry,i[1])
					if x:
						oid = prefix + " " + i[0]
						self.load(oid,x)
						listing[i[0]] = self.modules[oid]
				

		return listing

	def load(self,oid,module,pyfqdn="",path=""):
		module.namespace = oid.split()[0]
		module.prefix = oid

		###
		# override a path
		###
		if module.fbase:
			path = path.split("/")
			path[-1] = module.fbase
			path = "/".join(path)

		# ...
		if self.modules[module.namespace]["privileged"] != 1:
			if type(module.mask) != types.IntType:
				###
				# FIXME: give a warning!
				###
				module.mask = 0
			else:
				module.mask &= ~ flags.Esoteric

		if not self.modules.has_key(oid):
			self.modules[oid] = {
				"class": module,
				"path": path,
				"pyfqdn": pyfqdn,
			}

	def get_acquired(self,parent,name):

		bClass = parent["class"]
		base = bClass.prefix
		xClass = None

		## If we have a parent...
		if bClass:
			load = ""

			## try to get a class from it's export list
			if type(bClass.export) == types.ListType:
				for i in bClass.export:
					if type(i) == types.StringType:
						if i == name:
							load = name
					elif \
						(type(i) == types.ListType) or\
						(type(i) == types.TupleType):
						if i[0] == name:
							load = i[1]

			if load:
				try:
					xClass = getattr(bClass,load)
				except:
					return None

			## or look up in grand-parent :)
			if not xClass:
				return self.get_acquired(self.get(" ".join(base.split()[:-1])),name)

		return xClass


	def get_doc(self,oid):
		result = self.get(oid)["class"].doc
		if not result:
			result = inspect.getdoc(self.get(oid)["class"]())
		if not result:
			result = "<no description>"
		return result

def cfactory(pref):
	class cf(command):
		prefix = pref
		namespace = pref
		mask = flags.Begin | flags.Bypass

	return cf
