"""
ORB implementation core
"""

# 	Copyright (c) 2005-2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from threading import Thread
from cache import cache
from executive import service
from cxutil import flags
from cxutil.utils import fetch
import types
import os
import imp
import re

from auth import auth,taboo

from xml.dom.minidom import getDOMImplementation
imp = getDOMImplementation()


class orb(Thread):
	'''
	ORB server thread
	'''

	cache = None
	services = None
	map = None
	auth = None
	internal = None
	external = None
	wdir = None

	def __init__(self,bus,name="orb",wdir={}):
		Thread.__init__(self)
		self.setName(name)
		self.bus = bus
		self.auth = auth()
		self.cache = cache(self.bus,wdir)
		self.internal = {}
		self.external = {}
		self.services = {}
		self.map = {
			'get':			self.get,
			'cache':		self.magic_cache,
			'clean':		self.magic_clean,
			'getacl':		self.magic_getacl,
			'checkacl':		self.checkacl,
			'xmlDump':		self.xmlDump,
			'get_namespaces':	self.get_namespaces,
			'get_chain':		self.get_chain,
			'service':		self.service,
			'services':		self.magic_services,
			'register_service':	self.register_service,
			'register_module':	self.register_module,
			'register_acl':		self.register_acl,
			'unregister_acl':	self.unregister_acl,
			}



	def run(self):
		while True:
			addr = {}
			message = self.bus.get(addr)
			if type(message) == types.DictType:
				self.bus.put(addr["from"],self.map[message["map"]](message))
			if message == "shutdown":
				return

	def checkacl(self,opts):
		user = opts["creds"]["user"]
		group = opts["creds"]["group"]
		url = opts["url"]

		if not self.auth.check(user,"user",url):
			if not self.auth.check(group,"group",url):
				return False
		return True

	def get(self,opts):
		user = opts["creds"]["user"]
		group = opts["creds"]["group"]
		url = opts["prefix"] + " " + opts["name"]

		result = None
		if not self.checkacl({"creds": opts["creds"], "url": url}):
			return None

		r = self.cache.get(url)
		x = r["class"]
		if x:
			x = x()
		return x

	def magic_services(self,opts):
		'''
		Inspect services
		'''
		return self.services

	def magic_cache(self,opts):
		'''
		Inspect object cache
		'''
		return self.cache.bala()

	def magic_clean(self,opts):
		'''
		Clean object cache
		'''
		return self.cache.flush()

	def magic_getacl(self,opts):
		'''
		Inspect ACL state
		'''
		return self.auth.acl

	def register_service(self,opts):
		'''
		Register an external service

		opts['id'] - service id
		opts['path'] - path to service module (absolute or relative)
		'''
		name = opts["path"]
		xModule = None
		try:
			if name[-4:] == ".pyc":
				xModule = imp.load_compiled("runner.service",name)
			elif name[-3:] == ".py":
				xModule = imp.load_compiled("runner.service",name)

			if xModule:
				xClass = getattr(xModule,"service")
				xInstance = xClass()
			else:
				xInstance = service.proxy(name)

		except:
			return False

		self.services[opts["id"]] = xInstance
		return True

	def service(self,opts):
		'''
		'''
		if self.services.has_key(opts["id"]):
			return self.services[opts["id"]]
		else:
			return None

	def register_acl(self,opts):
		'''
		Register ACL entry
		'''
		return self.auth.register(opts["name"],opts["mode"],opts["path"])

	def unregister_acl(self,opts):
		'''
		Unregister ACL entry
		'''
		return self.auth.unregister(opts["name"],opts["mode"],opts["path"])

	def register_module(self,opts):
		'''
		Register an external module
		'''
		result = False
		try:
			sp = opts["sid"].split()
			base = " ".join(sp[:-1])
			name = sp[-1]
			result = self.cache.load(base,name,"",opts["xClass"],opts["xInstance"])
		except:
			pass
		if result:
			self.external[opts["sid"]] = opts["xInstance"]
		return result

	def add2list(self,list,namespace,name,item):
		if not list.has_key(namespace):
			list[namespace] = {}
		list[namespace][name] = item

	def get_namespaces(self,opts):
		return self.cache.wdir

	def get_chain(self,opts):
		'''
		Get commands chain from the command line
		'''

		creds = opts["creds"]
		text = opts["text"]

		namespace = fetch(opts,"namespace","")
		prefix = fetch(opts,"prefix",namespace)
		primary = fetch(opts,"primary",True)

		chain = []

		while True:

			first = None
			if text and prefix:
				## try to get the next word
				first = self.get_first_instance(creds,text,prefix)

			if first:
				## ok, we've got it, go on...
				chain.append(first)
				text = first[2]
				prefix = first[0].prefix
				## if there are parameters in a module:
				if first[0].parameters.keys():
					return chain

			else:
				## no such sybmbol; try another tree
				if primary:
					for (i,k,l) in self.get_namespaces(None):
						if i != namespace:
							c = self.get_chain({
								"creds": creds,
								"text": text,
								"namespace": i,
								"primary": False })
							if c:
								chain.extend(c)
								break

				return chain

	def get_first_instance(self,creds,text,prefix):
		'''
		Get the first instance for the command line
		'''
		text = text.strip()
		x = re.compile('^(\S+)(.*)',re.U)
		name = x.sub('\\1',text)
		rest = x.sub('\\2',text)

		n = name.split(":")
		if len(n) > 1:
			name = n[1]
			prefix = n[0]

		temp = self.get({
				"creds": creds,
				"name": name,
				"prefix": prefix,
			})
		if temp:
			return [temp,name,rest]
		else:
			return None

	def xmlDumpParameters(self,parameters,condition):

		dom = imp.createDocument(None, "parameters", None)
		root = dom.firstChild

		for (i,k) in parameters.items():

			node = dom.createElement("parameter")
			node.setAttribute("type",str(k["type"]))
			node.setAttribute("name",str(i))
			if k.has_key("pmask"):
				node.setAttribute("pmask",str(k["pmask"]))

			if k.has_key("range"):
				r = dom.createElement("range")
				for l in k["range"]:
					v = dom.createElement("v")
					v.appendChild(dom.createTextNode(str(l)))
					r.appendChild(v)
				node.appendChild(r)

			d = dom.createElement("desc")
			d.appendChild(dom.createTextNode(str(k["description"])))
			node.appendChild(d)

			if k.has_key("default"):
				d = dom.createElement("default")
				d.appendChild(dom.createTextNode(str(k["default"])))


			if k.has_key("condition"):
				if k["condition"](condition):
					root.appendChild(node)
			else:
				root.appendChild(node)
		return dom


	def xmlDump(self,opts):
		creds = opts["creds"]
		user = opts["creds"]["user"]
		group = opts["creds"]["group"]
		condition = opts["condition"]
		state = opts["state"]
		prefix = opts["prefix"]

		if not self.auth.check(user,"user",state):
			if not self.auth.check(group,"group",state):
				return imp.createDocument(None, "nodes", None)


		item = self.cache.get(state)
		path = item["path"]
		xInstance = item["class"]()
		parameters = xInstance.list()

		if len(parameters.keys()) and prefix:
			return self.xmlDumpParameters(parameters,condition)


		dom = imp.createDocument(None, "nodes", None)
		root = dom.firstChild


		try:
			root.setAttribute("ns",str(xInstance.namespace))

			for (i,k) in self.cache.get_unknown(item).items():
				if not self.checkacl({"creds": opts["creds"],"url": state + " " + i}):
					continue

				xInstance = k["class"]()
				if (xInstance.condition(condition)) and (not xInstance.mask & flags.Satellite):

					node = dom.createElement("node")
					node.setAttribute("name", str(i))
					node.setAttribute("mask", str(xInstance.mask))

					d = dom.createElement("desc")
					d.appendChild(dom.createTextNode(self.cache.get_doc(xInstance.prefix)))

					node.appendChild(d)
					root.appendChild(node)

					# get element's paramters description
					p = self.xmlDumpParameters(xInstance.list(),condition)
					node.appendChild(p.documentElement)

		except Exception,e:
			import traceback
			print e
			traceback.print_exc()
			pass

		return dom
