"""
Syntax checker
"""

# 	Copyright (c) 2005-2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import copy
import re
import types
from cxutil import pmask

def cmp_keys(x,y):
        if x[0] == '$' and y[0] == '$':
                x = x[1:]
                y = y[1:]
                return int(x) - int(y)
        elif x[0] == '$':
                return 1
        elif y[0] == '$':
                return -1
        else:
                if x > y:
                        return 1
                elif x < y:
                        return -1
                else:
                        return 0


class xSyntaxError(Exception):
	def __init__(self,description):
		self.description = description

	def __str__(self):
		return '%s' % (self.description)

class xSyntaxErrorNotCompleted(xSyntaxError):
	def __init__(self,description,param):
		xSyntaxError.__init__(self,description)
		self.param = param

class xSyntaxErrorNoMandatory(xSyntaxError):
	pass

class xParamError(Exception):
	def __init__(self,key,type,value):
		self.key = key
		self.value = value
		self.type = type

	def __str__(self):
		return 'parameter <%s> (type: %s; value: %s) type mismatch' % (self.key,self.type,self.value)


class oparser:
	'''
	'''

	parameters = None
	deprecated = None
	fname = None
	_keys = None
	re = {
		"ascii": re.compile("^[a-zA-Z0-9\._-]+$"),
		"email": re.compile("^[a-zA-Z0-9\._-]+\@([a-zA-Z0-9-]+\.)+[a-zA-Z]+$"),
		"int": re.compile("^[0-9]+$"),
		"interface": re.compile("^[0-9]+(\.[0-9]+)?$"),
		"ipaddr": re.compile("^([0-9]{,3}\.){3}[0-9]{,3}(/[0-9]{1,2})?$"),
		"speed": re.compile("^[0-9]+([KMGkmg]?bit)?$"),
		"size": re.compile("^[0-9]+([KMGkmg]?b?)?$"),
		"string": re.compile("^.+$"),
		"word": re.compile("^\S+$"),
	}

	## dict interface
	def keys(self):
		return self.parameters.keys()
	
	def items(self):
		return self.parameters.items()

	def has_key(self,key):
		return self.parameters.has_key(key)

	def __getitem__(self,key):
		return self.parameters[key]

	## ...
	def __init__(self,parameters,opts):
		'''
		'''
		self.deprecated = []
		self.parameters = {}
		self.fname = ""
		self._keys = parameters.keys()

		a = parameters.keys()
		a.sort(cmp_keys)
		opts = copy.copy(opts)
		positionals = []
		no_mandatory = None
		clean = False

		for i in a:

			if i[0] == "$" and not clean:
				clean = True
				for k in a:
					if k[0] != '$':
						self.remove_named(k,opts)


			if not parameters[i].has_key('mask'):
				parameters[i]["mask"] = pmask.Empty

			try:
				if (parameters[i]["mask"] & pmask.List) and (i[0] != "$"):
					x = []
					while True:
						try:
							(y,opts) = self.get(i,parameters[i]["type"],opts,positionals)
							x.append(y)
							self.remove_named(i,opts)
						except:
							break
					if not x:
						raise xSyntaxError("")
					else:
						self.parameters[i] = x
				else:
					(self.parameters[i],opts) = self.get(i,parameters[i]["type"],opts,positionals)

			except xSyntaxError,e:
				if parameters[i].has_key('default'):
					self.parameters[i] = parameters[i]['default']
				else:
					if not parameters[i]["mask"] & pmask.Optional:
						###
						# it is:
						# 1. mandatory...
						# 2. ... or defined with deprecated "optional" flag
						###
						if parameters[i].has_key('optional'):
							print parameters[i]
							self.deprecated.append((i,"`optional' flag will be removed in version 0.4.6"))
							if not parameters[i]['optional']:
								no_mandatory = xSyntaxErrorNoMandatory(str(e))
						else:
							raise e

			else:
				if parameters[i].has_key('range'):
					xRange = parameters[i]['range']
				else:
					xRange = []

				if (parameters[i]["mask"] & pmask.List) and (i[0] != "$"):
					x = self.parameters[i]
				else:
					x = [self.parameters[i],]

				for y in x:
					if self.check(y,parameters[i]["type"],xRange):
						if parameters[i]["type"] == 'enum':
							raise xParamError(i,"%s %s" % (parameters[i]["type"],xRange),y)
						else:
							raise xParamError(i,parameters[i]["type"],y)

				if (parameters[i]["mask"] & pmask.Fname) and (not parameters[i]["mask"] & pmask.List):
					self.fname = self.parameters[i]

		for k in a:
			if k[0] != '$':
				self.remove_named(k,opts)

		if positionals:
			opts = self.remove_positionals(positionals,opts)

		if opts:
			if (len(opts) == 1) and (opts[0] in self._keys):
				raise xSyntaxErrorNotCompleted('There must be value for %s' % (opts[0]),opts[0])
			else:
				raise xSyntaxError('parameter list too long, still remained %s' % (opts))
		
		if no_mandatory:
			raise no_mandatory


	def check(self,xParam,xType,xRange=[]):
		if xType == 'enum':
			# fast track parameter...
			if xParam in xRange:
				return None
			else:
				for i in xRange:
					if \
						(type(i) == types.TupleType) or \
						(type(i) == types.ListType):
							if self.re[i[1]].match(xParam):
								return None

		else:
			if not self.re.has_key(xType):
				raise xSyntaxError('unknown parameter type %s' % (xType))

			if self.re[xType].match(xParam):
				return None

		return type(xParam)

	def get(self,key,xType,opts,positionals):
		if key[0] == '$':
			positionals.append(int(key[1:]) - 1)
			return self.get_positional(int(key[1:]) - 1,xType,opts)
		else:
			return self.get_named(key,xType,opts)

	def remove_named(self,key,opts):
		if key in opts:
			if opts.index(key) < (len(opts) - 1):
				idx = opts.index(key)
				opts.pop(idx)
				opts.pop(idx)

	def remove_positionals(self,keys,opts):
		i = 0
		new_opts = []
		while i < len(opts):
			if not i in keys:
				new_opts.append(opts[i])
			i += 1
		return new_opts

	def get_positional(self,key,xType,opts):
		if key < len(opts):
			if xType == "string":
				result = " ".join(opts[key:])
				opts = opts[:key + 1]
				opts[key] = result
			else:
				result = opts[key]
			return (result,opts)

		raise xSyntaxError('parameter list too short (%s) to get $%i key' % (opts,key))


	def get_named(self,key,xType,opts):
		if key in opts:
			if opts.index(key) < (len(opts) - 1):
				if xType == "string":
					result = " ".join(opts[opts.index(key) + 1:])
					opts = opts[:opts.index(key) + 2]
					opts[opts.index(key) + 1] = result
				else:
					result = opts[opts.index(key) + 1]
				return (result,opts)

		raise xSyntaxError('no such parameter %s in %s' % (key,opts))
